//Segment - distance 2D
//Distance to a line segment.
//Translated from https://www.shadertoy.com/view/3tdSDj
//Original by Inigo Quilez

#![cfg_attr(target_arch = "spirv", no_std)]
#![feature(lang_items)]
#![feature(register_attr)]
#![register_attr(spirv)]

use spirv_std::glam::{Vec2, Vec3, Vec4};
use spirv_std::glam::{Vec2Swizzles, Vec3Swizzles, Vec4Swizzles};
use spirv_std::storage_class::{Input, Output, UniformConstant};

#[cfg(target_arch = "spirv")]
use spirv_std::num_traits::Float;

pub fn saturate(x: f32) -> f32 {
    x.max(0.0).min(1.0)
}

/// renamed because of cross-compilation issues with spirv-cross/ moltenvk
fn my_smoothstep(edge0: f32, edge1: f32, x: f32) -> f32 {
    // Scale, bias and saturate x to 0..1 range
    let x = saturate((x - edge0) / (edge1 - edge0));
    // Evaluate polynomial
    x * x * (3.0 - 2.0 * x)
}

// TODO: add this to glam? Rust std has it on f32/f64
fn pow(v: Vec3, power: f32) -> Vec3 {
    Vec3::new(v.x.powf(power), v.y.powf(power), v.z.powf(power))
}

// TODO: add this to glam? Rust std has it on f32/f64
fn exp(v: Vec3) -> Vec3 {
    Vec3::new(v.x.exp(), v.y.exp(), v.z.exp())
}

fn mix(x: f32, y: f32, a: f32) -> f32 {
    x * (1.0 - a) + y * a
}

fn mix_vec3(x: Vec3, y: Vec3, a: Vec3) -> Vec3 {
    x * (Vec3::splat(1.0) - a) + y * a
}

fn sign(v: f32) -> f32 {
    if v > 0.0 {
        1.0
    } else if v < 0.0 {
        -1.0
    } else {
        0.0
    }
}

fn ud_segment(p: Vec2, a: Vec2, b: Vec2) -> f32 {
    let ba = b - a;
    let pa = p - a;
    let h = (pa.dot(ba) / ba.dot(ba)).max(0.0).min(1.0);
    (pa - h * ba).length()
}

#[allow(unused_attributes)]
#[spirv(fragment)]

pub fn main_fs(
    #[spirv(frag_coord)] frag_coord: Input<Vec4>,
    u_time: UniformConstant<f32>,
    u_resolution: UniformConstant<Vec2>,
    mut output: Output<Vec4>,
) {
    let coord = frag_coord.load();
    let res = u_resolution.load();
    let time = u_time.load();
    let a = time.fract();

    let mut p = (2.0 * coord.xy() - res) / res.y;
    p *= 1.4;

    let v1 = Vec2::new((0.0 + time).cos(), (2.00 + time).cos());
    let v2 = Vec2::new((1.5 + time).cos(), (3.0 + time).cos());
    let th = 0.1 * (0.5 + 0.5 * (time * 1.1).sin());

    let d = ud_segment(p, v1, v2) - th;

    let mut col = Vec3::splat(1.0) - sign(d) * Vec3::new(0.1, 0.4, 0.7);
    col *= 1.0 - (-3.0 * d.abs()).exp();
    col *= 0.8 + 0.2 * (120.0 * d).cos();
    col = mix_vec3(
        col,
        Vec3::splat(1.0),
        Vec3::splat(1.0 - my_smoothstep(0.0, 0.015, d.abs())),
    );

    output.store(col.extend(1.0));
}
